/*
  Copyright (C) 2005 Sebastien Bourdeauducq

  This file is part of PictoSniff.

  PictoSniff is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  PictoSniff is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with PictoSniff ; if not, write to the Free Software Foundation,
  Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

#include <pthread.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <gtk/gtk.h>

#include <net/ethernet.h>
#include <net/if.h>
#include <net/if_media.h>

#include <net80211/ieee80211.h>
#include <net80211/ieee80211_ioctl.h>

#include "pictosniff.h"
#include "pictosniff.xpm"

static int set80211(int s, const char *iface, int type, int val, int len, u_int8_t *data)
{
  struct ieee80211req ireq;

  memset(&ireq, 0, sizeof(ireq));
  strncpy(ireq.i_name, iface, sizeof(ireq.i_name));
  ireq.i_type = type;
  ireq.i_val = val;
  ireq.i_len = len;
  ireq.i_data = data;
  if(ioctl(s, SIOCS80211, &ireq) < 0) return 0;
  return 1;
}

/*static int getmediaopt(int s, const char *iface, int options, int mode)
{
  struct ifmediareq ifmr;

  memset(&ifmr, 0, sizeof(ifmr));
  strncpy(ifmr.ifm_name, iface, sizeof(ifmr.ifm_name));

  if(ioctl(s, SIOCGIFMEDIA, &ifmr) < 0) {
    perror("cannot get ifmedia");
    return 0;
  }
  options = IFM_OPTIONS(ifmr.ifm_current);
  mode = IFM_MODE(ifmr.ifm_current);
  return 1;
}*/

static int setmediaopt(int s, const char *iface, int options, int mode)
{
  struct ifmediareq ifmr;
  struct ifreq ifr;
  int *mwords;

  memset(&ifmr, 0, sizeof(ifmr));
  strncpy(ifmr.ifm_name, iface, sizeof(ifmr.ifm_name));

  if(ioctl(s, SIOCGIFMEDIA, &ifmr) < 0) {
    perror("cannot get ifmedia");
    return 0;
  }
  if(ifmr.ifm_count == 0) {
    fputs("no media typed\n", stderr);
    return 0;
  }
  mwords = malloc(ifmr.ifm_count*sizeof(int));
  if(mwords == NULL) {
    perror("malloc");
    return 0;
  }
  ifmr.ifm_ulist = mwords;
  if(ioctl(s, SIOCGIFMEDIA, (caddr_t)&ifmr) < 0) {
    perror("cannot get ifmedia");
    return 0;
  }
  free(mwords);

  memset(&ifr, 0, sizeof(ifr));
  strncpy(ifr.ifr_name, iface, sizeof(ifr.ifr_name));
  ifr.ifr_media = (ifmr.ifm_current &~ IFM_OMASK) | options;
  ifr.ifr_media = (ifr.ifr_media &~ IFM_MMASK) | IFM_MAKEMODE(mode);

  if(ioctl(s, SIOCSIFMEDIA, (caddr_t)&ifr) < 0) {
    perror("cannot set ifmedia");
    return 0;
  }
  return 1;
}

static int getifflags(int s, const char *iface, int *flags)
{
  struct ifreq ifr;

  strncpy(ifr.ifr_name, iface, sizeof (ifr.ifr_name));
  if(ioctl(s, SIOCGIFFLAGS, &ifr) < 0) {
    perror("SIOCGIFFLAGS ioctl failed");
    return 0;
  }
  *flags = (ifr.ifr_flags & 0xffff)|(ifr.ifr_flagshigh << 16);
  return 1;
}

static int setifflags(int s, const char *iface, int flags)
{
  struct ifreq ifr;

  memset(&ifr, 0, sizeof(ifr));
  strncpy(ifr.ifr_name, iface, sizeof(ifr.ifr_name));
  ifr.ifr_flags = flags & 0xffff;
  ifr.ifr_flagshigh = flags >> 16;
  if(ioctl(s, SIOCSIFFLAGS, &ifr) < 0) {
    perror("SIOCSIFFLAGS ioctl failed");
    return 0;
  }
  return 1;
}

enum {
  COL_TIMESTAMP = 0,
  COL_MESSAGE,
  NUM_COLS
};

static GtkWidget *main_window;

static GtkWidget *vbox;
static GtkWidget *hbox;

static GtkWidget *treescroll;
static GtkWidget *treeview;
static GtkListStore *treestore;
static GtkTreeIter treeiter;

static GtkWidget *interface_label;
static GtkWidget *interface_separator;
static GtkWidget *channel_label;
static GtkWidget *channel_spin;
static GtkWidget *channel_separator;
static GtkWidget *rssi_label;
static GtkWidget *rssi_bar;
static GtkWidget *rssi_separator;
static GtkWidget *save_button;
static GtkWidget *clear_button;

static const char *interface;
static int ioctl_socket;

static void add_pictosniff_banner()
{
  GdkPixbuf *pixbuf;
  time_t t;
  char *t2, *c;
  
  pixbuf = gdk_pixbuf_new_from_xpm_data(&pictosniff_xpm[0]);

  gtk_list_store_append(treestore, &treeiter);
  time(&t);
  t2 = ctime(&t);
  c = strchr(t2, '\n');
  if(c != NULL) *c = 0;
  gtk_list_store_set(treestore, &treeiter, COL_TIMESTAMP, t2, COL_MESSAGE, pixbuf, -1);
  
  /* pixbuf has a refcount of 2 now, as the list store has added its own reference */
  g_object_unref(pixbuf);
}

static GtkTreeModel *create_and_fill_model()
{
  treestore = gtk_list_store_new(NUM_COLS, G_TYPE_STRING, GDK_TYPE_PIXBUF);
  add_pictosniff_banner();
  return GTK_TREE_MODEL(treestore);
}

static GtkWidget *create_view_and_model()
{
  GtkTreeViewColumn *col;
  GtkCellRenderer *renderer;
  GtkTreeModel *model;
  GtkWidget *view;

  view = gtk_tree_view_new();

  /* --- Column #1 --- */

  renderer = gtk_cell_renderer_text_new();
  gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(view), -1, "Time", renderer, "text", COL_TIMESTAMP, NULL);

  /* --- Column #2 --- */

  col = gtk_tree_view_column_new();

  renderer = gtk_cell_renderer_pixbuf_new();
  gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW (view), -1, "Message", renderer, "pixbuf", COL_MESSAGE, NULL);

  model = create_and_fill_model();

  gtk_tree_view_set_model(GTK_TREE_VIEW(view), model);

  g_object_unref(model); /* destroy model automatically with view */
  
  gtk_tree_selection_set_mode(gtk_tree_view_get_selection(GTK_TREE_VIEW(view)), GTK_SELECTION_NONE);

  return view;
}

static void pixbuf_free(guchar *pixels, gpointer data)
{
  free(pixels);
}

int add_message(const char *data, int w, int h)
{
  GdkPixbuf *pixbuf;
  time_t t;
  char *t2, *c;
  //GtkAdjustment* adj;
  

  gdk_threads_enter();
  pixbuf = gdk_pixbuf_new_from_data(data, GDK_COLORSPACE_RGB, FALSE, 8, w, h, w*3, pixbuf_free, NULL);

  gtk_list_store_append(treestore, &treeiter);
  time(&t);
  t2 = ctime(&t);
  c = strchr(t2, '\n');
  if(c != NULL) *c = 0;
  gtk_list_store_set(treestore, &treeiter, COL_TIMESTAMP, t2, COL_MESSAGE, pixbuf, -1);
  
  /* pixbuf has a refcount of 2 now, as the list store has added its own reference */
  g_object_unref(pixbuf);
  /* scroll down */
  /*adj = gtk_scrolled_window_get_vadjustment(GTK_SCROLLED_WINDOW(treescroll));
  printf("lower = %f upper = %f ps = %f\n", adj->lower, adj->upper, adj->page_size);
  gtk_adjustment_set_value(adj, adj->page_size-adj->upper);
  */
  gdk_threads_leave();
  
  return 1;
}

int update_rssi(int rssi_level)
{
  gdk_threads_enter();
  gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(rssi_bar), rssi_level > MAX_RSSI ? 1.0 : ((gdouble)rssi_level)/((gdouble)MAX_RSSI));
  gdk_threads_leave();
  return 1;
}

static void channel_change(GtkWidget *spinbutton, gpointer user_data)
{
  int c;
  
  c = (int)gtk_spin_button_get_value(GTK_SPIN_BUTTON(channel_spin));
  if(!set80211(ioctl_socket, interface, IEEE80211_IOC_CHANNEL, c, 0, NULL)) fprintf(stderr, "failed to set channel %d", c);
}

static void clear_button_press(GtkWidget *widget, gpointer data)
{
  gtk_list_store_clear(treestore);
  add_pictosniff_banner();
}

static void save_button_press(GtkWidget *widget, gpointer data)
{
  GtkWidget *dialog;
  gchar *basename;
  gboolean valid;
  GtkTreeIter iter;
  gchar *timestamp;
  GdkPixbuf *pixbuf;
  int i;
  gchar *c;
  char *filename;

  dialog = gtk_file_chooser_dialog_new("Save to file", GTK_WINDOW(main_window), GTK_FILE_CHOOSER_ACTION_SAVE, GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL, GTK_STOCK_SAVE, GTK_RESPONSE_ACCEPT, NULL);
  if(gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_ACCEPT) {
    basename = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(dialog));
    valid = gtk_tree_model_get_iter_first(GTK_TREE_MODEL(treestore), &iter);
    i = 0;
    while(valid) {
      if(i > 0) { /* Don't save the PictoSniff banner */
        gtk_tree_model_get(GTK_TREE_MODEL(treestore), &iter, COL_TIMESTAMP, &timestamp, COL_MESSAGE, &pixbuf, -1);
        c = timestamp;
        while(*c != 0) {
          if(*c == ' ') *c = '_';
          if(*c == ':') *c = '_';
          c++;
        }
        if(asprintf(&filename, "%s_%04d_%s.png", basename, i, timestamp) == -1) {
          g_free(timestamp);
          perror("asprintf");
          break;
        }
        g_free(timestamp);
        gdk_pixbuf_save(pixbuf, filename, "png", NULL, NULL);
        free(filename);
        g_object_unref(pixbuf);
      }
      valid = gtk_tree_model_iter_next(GTK_TREE_MODEL(treestore), &iter);
      i++;
    }
    g_free(basename);
  }
  gtk_widget_destroy(dialog); 
}

static gboolean delete_event(GtkWidget *widget, GdkEvent *event, gpointer data)
{
  return FALSE;
}

static void destroy(GtkWidget *widget, gpointer data)
{
  gtk_main_quit();
}

int main(int argc, char *argv[])
{
  int prev_flags;
  
  g_thread_init(NULL);
  gdk_threads_init();
  gdk_threads_enter();
  gtk_init(&argc, &argv);
  
  if(argc != 2) {
    fputs("Usage : pictosniff <network interface>\n", stderr);
    gtk_exit(1);
  }
  
  interface = argv[1];
  /* Set up the wireless interface */
  if((ioctl_socket = socket(AF_INET, SOCK_DGRAM, 0)) == -1) {
    perror("socket");
    gtk_exit(1);
  }
  if(!setmediaopt(ioctl_socket, interface, IFM_IEEE80211_MONITOR, IFM_AUTO)) {
    fputs("Unable to enter monitor mode\n", stderr);
    gtk_exit(1);
  }
  if(!set80211(ioctl_socket, interface, IEEE80211_IOC_CHANNEL, 1, 0, NULL)) {
    fputs("failed to set initial channel\n", stderr);
    gtk_exit(1);
  }
  if(!getifflags(ioctl_socket, interface, &prev_flags)) {
    fputs("failed to get interface flags\n", stderr);
    gtk_exit(1);
  }
  if(!setifflags(ioctl_socket, interface, prev_flags|IFF_PPROMISC|IFF_UP)) {
    fputs("failed to set interface flags\n", stderr);
    gtk_exit(1);
  }

  main_window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title(GTK_WINDOW(main_window), "PictoSniff!");
  g_signal_connect(G_OBJECT(main_window), "delete_event", G_CALLBACK (delete_event), NULL);
  g_signal_connect(G_OBJECT(main_window), "destroy", G_CALLBACK(destroy), NULL);
  gtk_container_set_border_width(GTK_CONTAINER(main_window), 10);
  gtk_window_set_default_size(GTK_WINDOW(main_window), 490, 470);
  
  vbox = gtk_vbox_new(FALSE, 1);
  gtk_container_add(GTK_CONTAINER(main_window), vbox);
  
  treescroll = gtk_scrolled_window_new(NULL, NULL);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(treescroll), GTK_POLICY_NEVER, GTK_POLICY_AUTOMATIC);
  treeview = create_view_and_model();
  gtk_container_add(GTK_CONTAINER(treescroll), treeview);
  gtk_box_pack_start(GTK_BOX(vbox), treescroll, TRUE, TRUE, 0);
  
  hbox = gtk_hbox_new(FALSE, 2);
  gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);
  
  interface_label = gtk_label_new(interface);
  gtk_box_pack_start(GTK_BOX(hbox), interface_label, FALSE, FALSE, 0);
  interface_separator = gtk_vseparator_new();
  gtk_box_pack_start(GTK_BOX(hbox), interface_separator, FALSE, FALSE, 0);
  channel_label = gtk_label_new("Channel :");
  gtk_box_pack_start(GTK_BOX(hbox), channel_label, FALSE, FALSE, 0);
  channel_spin = gtk_spin_button_new_with_range(1.0, 14.0, 1.0);
  gtk_box_pack_start(GTK_BOX(hbox), channel_spin, FALSE, FALSE, 0);
  channel_separator = gtk_vseparator_new();
  gtk_box_pack_start(GTK_BOX(hbox), channel_separator, FALSE, FALSE, 0);
  rssi_label = gtk_label_new("RSSI :");
  gtk_box_pack_start(GTK_BOX(hbox), rssi_label, FALSE, FALSE, 0);
  rssi_bar = gtk_progress_bar_new();
  gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(rssi_bar), 0.0);
  gtk_box_pack_start(GTK_BOX(hbox), rssi_bar, FALSE, FALSE, 0);
  rssi_separator = gtk_vseparator_new();
  gtk_box_pack_start(GTK_BOX(hbox), rssi_separator, FALSE, FALSE, 0);
  save_button = gtk_button_new_from_stock(GTK_STOCK_SAVE);
  gtk_box_pack_start(GTK_BOX(hbox), save_button, FALSE, FALSE, 0);
  clear_button = gtk_button_new_from_stock(GTK_STOCK_CLEAR);
  gtk_box_pack_start(GTK_BOX(hbox), clear_button, FALSE, FALSE, 0);
  
  g_signal_connect(G_OBJECT(channel_spin), "value-changed", G_CALLBACK(channel_change), NULL);
  g_signal_connect(G_OBJECT(clear_button), "clicked", G_CALLBACK(clear_button_press), NULL);
  g_signal_connect(G_OBJECT(save_button), "clicked", G_CALLBACK(save_button_press), NULL);
  
  gtk_widget_show_all(main_window);
  
  if(!start_sniffing(interface)) gtk_exit(1);
  gtk_main();
  end_sniffing();
  
  close(ioctl_socket);
  
  gdk_threads_leave();

  return 0;
}
